#include <ftl/cuda_common.hpp>
#include <ftl/cuda_util.hpp>
#include <ftl/depth_camera.hpp>
#include "depth_camera_cuda.hpp"

#define T_PER_BLOCK 16
#define MINF __int_as_float(0xff800000)

using ftl::voxhash::DepthCameraCUDA;
using ftl::voxhash::HashData;
using ftl::voxhash::HashParams;

extern __constant__ ftl::voxhash::DepthCameraCUDA c_cameras[MAX_CAMERAS];

/// ===== MLS Smooth

// TODO:(Nick) Put this in a common location (used in integrators.cu)
extern __device__ float spatialWeighting(float r);

#define WINDOW_RADIUS 5

__global__ void mls_smooth_kernel(ftl::cuda::TextureObject<float> output, HashData hashData, HashParams hashParams, int numcams, int cam) {
	const unsigned int x = blockIdx.x*blockDim.x + threadIdx.x;
	const unsigned int y = blockIdx.y*blockDim.y + threadIdx.y;

	const int width = output.width();
	const int height = output.height();

	const DepthCameraCUDA &mainCamera = c_cameras[cam];

	if (x < width && y < height) {
		output(x,y) = 1000.0f;
		__syncthreads();

		float depth = tex2D<float>(mainCamera.depth2, x, y);

		float3 wpos = make_float3(0.0f);
		float3 wnorm = make_float3(0.0f);
		float weights = 0.0f;

		if (depth >= mainCamera.params.m_sensorDepthWorldMin && depth <= mainCamera.params.m_sensorDepthWorldMax) {
			const float3 mPos = mainCamera.pose * mainCamera.params.kinectDepthToSkeleton(x, y, depth);

			for (uint cam=0; cam<numcams; ++cam) {
				const ftl::voxhash::DepthCameraCUDA &camera = c_cameras[cam];
				const uint height = camera.params.m_imageHeight;
				const uint width = camera.params.m_imageWidth;
			
				const float3 pf = camera.poseInverse * mPos;
				const uint2 screenPos = make_uint2(camera.params.cameraToKinectScreenInt(pf));

				#pragma unroll
				for (int v=-WINDOW_RADIUS; v<=WINDOW_RADIUS; ++v) {
					for (int u=-WINDOW_RADIUS; u<=WINDOW_RADIUS; ++u) {
						if (screenPos.x+u < width && screenPos.y+v < height) {	//on screen
							float depth = tex2D<float>(camera.depth2, screenPos.x+u, screenPos.y+v);

							//float4 normal = tex2D<float4>(camera.normal, screenPos.x+u, screenPos.y+v);
							const float3 worldPos = camera.pose * camera.params.kinectDepthToSkeleton(screenPos.x+u, screenPos.y+v, depth);
							const float weight = spatialWeighting(length(mPos - worldPos));

							wpos += weight*worldPos;
							//wnorm += weight*make_float3(normal);
							weights += weight;				
						}
					}
				}
			}

			wpos /= weights;
		}

		wpos = mainCamera.poseInverse * wpos;
		const uint2 screenPos = make_uint2(mainCamera.params.cameraToKinectScreenInt(wpos));
		if (screenPos.x < output.width() && screenPos.y < output.height()) {
			output(screenPos.x,screenPos.y) = (weights >= hashParams.m_confidenceThresh) ? wpos.z : 1000.0f;
			//output(x,y) = (weights >= hashParams.m_confidenceThresh) ? wpos.z : 1000.0f;
		}
	}
}

void ftl::cuda::mls_smooth(TextureObject<float> &output, const HashData &hashData, const HashParams &hashParams, int numcams, int cam, cudaStream_t stream) {
	const dim3 gridSize((output.width() + T_PER_BLOCK - 1)/T_PER_BLOCK, (output.height() + T_PER_BLOCK - 1)/T_PER_BLOCK);
	const dim3 blockSize(T_PER_BLOCK, T_PER_BLOCK);

	mls_smooth_kernel<<<gridSize, blockSize, 0, stream>>>(output, hashData, hashParams, numcams, cam);

#ifdef _DEBUG
	cudaSafeCall(cudaDeviceSynchronize());
#endif
}


/// ===== Point cloud from depth =====

__global__ void point_cloud_kernel(float3* output, DepthCameraCUDA depthCameraData)
{
	const unsigned int x = blockIdx.x*blockDim.x + threadIdx.x;
	const unsigned int y = blockIdx.y*blockDim.y + threadIdx.y;

	const int width = depthCameraData.params.m_imageWidth;
	const int height = depthCameraData.params.m_imageHeight;

	if (x < width && y < height) {
		float depth = tex2D<float>(depthCameraData.depth, x, y);

		output[y*width+x] = (depth >= depthCameraData.params.m_sensorDepthWorldMin && depth <= depthCameraData.params.m_sensorDepthWorldMax) ?
			depthCameraData.params.kinectDepthToSkeleton(x, y, depth) :
			make_float3(MINF, MINF, MINF);
	}
}

void ftl::cuda::point_cloud(float3* output, const DepthCameraCUDA &depthCameraData, cudaStream_t stream) {
	const dim3 gridSize((depthCameraData.params.m_imageWidth + T_PER_BLOCK - 1)/T_PER_BLOCK, (depthCameraData.params.m_imageHeight + T_PER_BLOCK - 1)/T_PER_BLOCK);
	const dim3 blockSize(T_PER_BLOCK, T_PER_BLOCK);

	point_cloud_kernel<<<gridSize, blockSize, 0, stream>>>(output, depthCameraData);

#ifdef _DEBUG
	cudaSafeCall(cudaDeviceSynchronize());
#endif
}

/// ===== NORMALS =====


__global__ void compute_normals_kernel(const float3 *input, ftl::cuda::TextureObject<float4> output)
{
	const unsigned int x = blockIdx.x*blockDim.x + threadIdx.x;
	const unsigned int y = blockIdx.y*blockDim.y + threadIdx.y;

	const int width = output.width();

	if(x >= output.width() || y >= output.height()) return;

	output(x,y) = make_float4(MINF, MINF, MINF, MINF);

	if(x > 0 && x < output.width()-1 && y > 0 && y < output.height()-1)
	{
		// TODO:(Nick) Should use a 7x7 window
		const float3 CC = input[(y+0)*width+(x+0)];
		const float3 PC = input[(y+1)*width+(x+0)];
		const float3 CP = input[(y+0)*width+(x+1)];
		const float3 MC = input[(y-1)*width+(x+0)];
		const float3 CM = input[(y+0)*width+(x-1)];

		if(CC.x != MINF && PC.x != MINF && CP.x != MINF && MC.x != MINF && CM.x != MINF)
		{
			const float3 n = cross(PC-MC, CP-CM);
			//const float  l = length(n);

			//if(l > 0.0f)
			//{
				output(x,y) = make_float4(n, 1.0f); //make_float4(n/-l, 1.0f);
			//}
		}
	}
}

void ftl::cuda::compute_normals(const float3 *input, ftl::cuda::TextureObject<float4> *output, cudaStream_t stream) {
	const dim3 gridSize((output->width() + T_PER_BLOCK - 1)/T_PER_BLOCK, (output->height() + T_PER_BLOCK - 1)/T_PER_BLOCK);
	const dim3 blockSize(T_PER_BLOCK, T_PER_BLOCK);

	compute_normals_kernel<<<gridSize, blockSize, 0, stream>>>(input, *output);

#ifdef _DEBUG
	cudaSafeCall(cudaDeviceSynchronize());
	//cutilCheckMsg(__FUNCTION__);
#endif
}